exception LoadRawImg_error of string
exception LoadRaw3DS_error of string

let input_int ic =
	let c1 = input_byte ic 
	and c2 = input_byte ic 
	and c3 = input_byte ic 
	and c4 = input_byte ic in
	(c4 lsl 12) + (c3 lsl 8) + (c2 lsl 4) + c1
;;

let loadRawImg fname =
	try
	let ic 	= open_in_bin fname in
	let width  = input_int ic 
	and height = input_int ic in
	let len = width * (height lsl 2) in
	let data = (String.make len (char_of_int 0)) in
		let rec loop l =
			if l < len then begin
				let c = input_byte ic in 
					data.[l] <- char_of_int c;
					loop (l + 1);
			end;
		in
		loop 0;		

		(width, height, data)
	with _ -> raise (LoadRawImg_error fname)		
;;



let loadRaw3DS fname =
	try
	let ic 	= open_in_bin fname in
	let num = (in_channel_length ic) in
	let farray = Array.make (num/8) 0.0 in
	let ms  = Marshal.to_string farray [] in
	let offset = (String.length ms) - num in
	let s = String.make num (char_of_int 0) in
		let rec loop l =
			if l < num then begin
				let c = input_byte ic in 
					s.[l] <-  char_of_int c;
					loop (l + 1);
			end;
		in
		loop 0;			
		String.blit s 0 ms offset num;
		let ret = Marshal.from_string ms 0 in
		ret
	with
	_ -> raise (LoadRaw3DS_error fname)
;;

(*
try
	let s = loadRawImg "data/base.jpeg.raw" in ()
with
LoadRawImg_error s -> Printf.printf "Unable to load %s" s;;	

try
	let t = loadRaw3DS "data/model.3ds.raw" in 
	Array.iter (Printf.printf "%f\n") t 
with
LoadRaw3DS_error t -> Printf.printf "Unable to load %s" t;;	
*)
