(*  3d math lib
 *
 *.      written by Alexander Zaprjagaev
 *.      frustum@public.tsu.ru
 *.      2:5005/93.15@FidoNet
 *)

let lessThanEpsilon x =
 ((abs_float x) < epsilon_float);;
 
let pi = 3.14159265358979323846;;

let deg2rad = (pi /. 180.0);;

let rad2deg = (180.0 /. pi);;
 
let vectorLength v =
     sqrt (v.(0) *. v.(0) +. v.(1) *. v.(1) +. v.(2) *. v.(2));;

let vectorScale i v =
 [| i *. v.(0); i *. v.(1); i *. v.(2)|];;     

let vectorAdd a b = [|a.(0) +. b.(0); a.(1) +. b.(1); a.(2) +. b.(2) |];;

let vectorSub a b = [|a.(0) -. b.(0); a.(1) -. b.(1); a.(2) -. b.(2) |];;

let vectorDotProduct a b = a.(0) *. b.(0) +. a.(1) *. b.(1) +. a.(2) *. b.(2);; 
 
let vectorNormalize v =
    let length = sqrt(v.(0) *. v.(0) +. v.(1) *. v.(1) +. v.(2) *. v.(2)) in
    if (lessThanEpsilon length) then 
        ([| 0.0; 0.0; 0.0 |], 0.0)
	else   
     let i = 1.0 /. length in
    	( (vectorScale i v), length);;    		

let vectorCrossProduct v1 v2 =
    [| v1.(1) *. v2.(2) -. v1.(2) *. v2.(1);
       v1.(2) *. v2.(0) -. v1.(0) *. v2.(2);
       v1.(0) *. v2.(1) -. v1.(1) *. v2.(0) |];;

let vectorTransform v1 m =
[|  v1.(0) *. m.(0) +. v1.(1) *. m.(4) +. v1.(2) *. m.(8) +. m.(12);
    v1.(0) *. m.(1) +. v1.(1) *. m.(5) +. v1.(2) *. m.(9) +. m.(13);
    v1.(0) *. m.(2) +. v1.(1) *. m.(6) +. v1.(2) *. m.(10) +. m.(14) |];;

let vectorTransformNormal v1 m =
 [| v1.(0) *. m.(0) +. v1.(1) *. m.(4) +. v1.(2) *. m.(8);
    v1.(0) *. m.(1) +. v1.(1) *. m.(5) +. v1.(2) *. m.(9);
    v1.(0) *. m.(2) +. v1.(1) *. m.(6) +. v1.(2) *. m.(10) |];;

let matrixMultiply m1 m2 =
  [|m1.(0) *. m2.(0) +. m1.(4) *. m2.(1) +. m1.(8) *. m2.(2) +. m1.(12) *. m2.(3);
    m1.(1) *. m2.(0) +. m1.(5) *. m2.(1) +. m1.(9) *. m2.(2) +. m1.(13) *. m2.(3);
    m1.(2) *. m2.(0) +. m1.(6) *. m2.(1) +. m1.(10) *. m2.(2) +. m1.(14) *. m2.(3);
    m1.(3) *. m2.(0) +. m1.(7) *. m2.(1) +. m1.(11) *. m2.(2) +. m1.(15) *. m2.(3);
    m1.(0) *. m2.(4) +. m1.(4) *. m2.(5) +. m1.(8) *. m2.(6) +. m1.(12) *. m2.(7);
    m1.(1) *. m2.(4) +. m1.(5) *. m2.(5) +. m1.(9) *. m2.(6) +. m1.(13) *. m2.(7);
    m1.(2) *. m2.(4) +. m1.(6) *. m2.(5) +. m1.(10) *. m2.(6) +. m1.(14) *. m2.(7);
    m1.(3) *. m2.(4) +. m1.(7) *. m2.(5) +. m1.(11) *. m2.(6) +. m1.(15) *. m2.(7);
    m1.(0) *. m2.(8) +. m1.(4) *. m2.(9) +. m1.(8) *. m2.(10) +. m1.(12) *. m2.(11);
    m1.(1) *. m2.(8) +. m1.(5) *. m2.(9) +. m1.(9) *. m2.(10) +. m1.(13) *. m2.(11);
    m1.(2) *. m2.(8) +. m1.(6) *. m2.(9) +. m1.(10) *. m2.(10) +. m1.(14) *. m2.(11);
    m1.(3) *. m2.(8) +. m1.(7) *. m2.(9) +. m1.(11) *. m2.(10) +. m1.(15) *. m2.(11);
    m1.(0) *. m2.(12) +. m1.(4) *. m2.(13) +. m1.(8) *. m2.(14) +. m1.(12) *. m2.(15);
    m1.(1) *. m2.(12) +. m1.(5) *. m2.(13) +. m1.(9) *. m2.(14) +. m1.(13) *. m2.(15);
    m1.(2) *. m2.(12) +. m1.(6) *. m2.(13) +. m1.(10) *. m2.(14) +. m1.(14) *. m2.(15);
    m1.(3) *. m2.(12) +. m1.(7) *. m2.(13) +. m1.(11) *. m2.(14) +. m1.(15) *. m2.(15) |];;    

let matrixInverse m =
    let   invdet =
    m.(0) *. m.(5) *. m.(10)
    +. m.(4) *. m.(9) *. m.(2)
    +. m.(8) *. m.(1) *. m.(6)
    -. m.(8) *. m.(5) *. m.(2)
    -. m.(4) *. m.(1) *. m.(10)
    -. m.(0) *. m.(9) *. m.(6) in
    if (invdet *. invdet < 1e-25) then (false, Array.make 16 0.0)
    else
    let det = 1.0 /. invdet in  
    let out = Array.make 16 0.0 in
    out.(0) <-    (m.(5) *. m.(10) -. m.(9) *. m.(6)) *. det;
    out.(1) <-  -. (m.(1) *. m.(10) -. m.(9) *. m.(2)) *. det;
    out.(2) <-    (m.(1) *. m.(6) -.  m.(5) *. m.(2)) *. det;
    out.(3) <- 0.0;
    out.(4) <-  -. (m.(4) *. m.(10) -. m.(8) *. m.(6)) *. det;
    out.(5) <-    (m.(0) *. m.(10) -. m.(8) *. m.(2)) *. det;
    out.(6) <-  -. (m.(0) *. m.(6) -.  m.(4) *. m.(2)) *. det;
    out.(7) <- 0.0;
    out.(8) <-    (m.(4) *. m.(9) -.  m.(8) *. m.(5)) *. det;
    out.(9) <-  -. (m.(0) *. m.(9) -.  m.(8) *. m.(1)) *. det;
    out.(10) <-   (m.(0) *. m.(5) -.  m.(4) *. m.(1)) *. det;
    out.(11) <- 0.0;
    out.(12) <- -. (m.(12) *. out.(0) +. m.(13) *. out.(4) +. m.(14) *. out.(8));
    out.(13) <- -. (m.(12) *. out.(1) +. m.(13) *. out.(5) +. m.(14) *. out.(9));
    out.(14) <- -. (m.(12) *. out.(2) +. m.(13) *. out.(6) +. m.(14) *. out.(10));
    out.(15) <- 1.0;  
    (true, out);;

let matrixTranspose m =
	let out = Array.make 16 0.0 in	
	out.(0) <- m.(0); out.(4) <- m.(1); out.(8) <- m.(2); out.(12) <- m.(3);
	out.(1) <- m.(4); out.(5) <- m.(5); out.(9) <- m.(6); out.(13) <- m.(7);
	out.(2) <- m.(8); out.(6) <- m.(9); out.(10) <- m.(10); out.(14) <- m.(11);
	out.(3) <- m.(12); out.(7) <- m.(13); out.(11) <- m.(14); out.(15) <- m.(15);
	out;;

let matrixRotationX angle =
    let rad = angle *. deg2rad in
    let cCos = cos(rad)
    and cSin = sin(rad) 
    and out = Array.make 16 0.0 in
    out.(0) <- 1.0; out.(4) <- 0.0; out.(8) <- 0.0; out.(12) <- 0.0;
    out.(1) <- 0.0; out.(5) <- cCos; out.(9) <- -.cSin; out.(13) <- 0.0;
    out.(2) <- 0.0; out.(6) <- cSin; out.(10) <- cCos; out.(14) <- 0.0;
    out.(3) <- 0.0; out.(7) <- 0.0; out.(11) <- 0.0; out.(15) <- 1.0;
	out;;

let matrixRotationY angle =
    let rad = angle *. deg2rad in
    let cCos = cos(rad)
    and cSin = sin(rad) 
    and out = Array.make 16 0.0 in
    out.(0) <- cCos; out.(4) <- 0.0; out.(8) <- cSin; out.(12) <- 0.0;
    out.(1) <- 0.0; out.(5) <- 1.0; out.(9) <- 0.0; out.(13) <- 0.0;
    out.(2) <- -.cSin; out.(6) <- 0.0; out.(10) <- cCos; out.(14) <- 0.0;
    out.(3) <- 0.0; out.(7) <- 0.0; out.(11) <- 0.0; out.(15) <- 1.0;
	out;;

let matrixRotationZ angle =
    let rad = angle *. deg2rad in
    let cCos = cos(rad)
    and cSin = sin(rad) 
    and out = Array.make 16 0.0 in    
    out.(0) <- cCos; out.(4) <- -.cSin; out.(8) <- 0.0; out.(12) <- 0.0;
    out.(1) <- cSin; out.(5) <- cCos; out.(9) <- 0.0; out.(13) <- 0.0;
    out.(2) <- 0.0; out.(6) <- 0.0; out.(10) <- 1.0; out.(14) <- 0.0;
    out.(3) <- 0.0; out.(7) <- 0.0; out.(11) <- 0.0; out.(15) <- 1.0;
	out;;

let matrixTranslate x y z =
    let out = Array.make 16 0.0 in    
    out.(0) <- 1.0; out.(4) <- 0.0; out.(8) <- 0.0; out.(12) <- x;
    out.(1) <- 0.0; out.(5) <- 1.0; out.(9) <- 0.0; out.(13) <- y;
    out.(2) <- 0.0; out.(6) <- 0.0; out.(10) <- 1.0; out.(14) <- z;
    out.(3) <- 0.0; out.(7) <- 0.0; out.(11) <- 0.0; out.(15) <- 1.0;
	out;;

let matrixLookAt eye dir up =
    let z, _ = vectorNormalize (vectorSub eye dir) in
    let x, _ = vectorNormalize (vectorCrossProduct up z) in
    let y, _ = vectorNormalize (vectorCrossProduct z x) 
	and m1 = Array.make 16 0.0 in
    m1.(0) <- x.(0); m1.(4) <- x.(1); m1.(8) <- x.(2); m1.(12) <- 0.0;
    m1.(1) <- y.(0); m1.(5) <- y.(1); m1.(9) <- y.(2); m1.(13) <- 0.0;
    m1.(2) <- z.(0); m1.(6) <- z.(1); m1.(10) <- z.(2); m1.(14) <- 0.0;
    m1.(3) <- 0.0; m1.(7) <- 0.0; m1.(11) <- 0.0; m1.(15) <- 1.0;
    let m2 = matrixTranslate (-.eye.(0)) (-.eye.(1)) (-.eye.(2)) in
    matrixMultiply m1 m2;;

let matrixShadow plane light =
    let dot = plane.(0) *. light.(0) +. plane.(1) *. light.(1) +.
          plane.(2) *. light.(2) +. plane.(3) *. light.(3) in
    [| -.light.(0) *. plane.(0) +. dot;
     -.light.(1) *. plane.(0);
     -.light.(2) *. plane.(0);
     -.light.(3) *. plane.(0);
     -.light.(0) *. plane.(1);
     -.light.(1) *. plane.(1) +. dot;
     -.light.(2) *. plane.(1);
     -.light.(3) *. plane.(1);
     -.light.(0) *. plane.(2);
     -.light.(1) *. plane.(2);
     -.light.(2) *. plane.(2) +. dot;
     -.light.(3) *. plane.(2);
     -.light.(0) *. plane.(3);
     -.light.(1) *. plane.(3);
     -.light.(2) *. plane.(3);
     -.light.(3) *. plane.(3) +. dot |];;

let quaternionSet dir angle =
    let length = sqrt(dir.(0) *. dir.(0) +. dir.(1) *. dir.(1) +. dir.(2) *. dir.(2)) in
    if( lessThanEpsilon length) then
        let ilength = 1.0 /. length and
        a = angle *. deg2rad /. 2.0 in
        let sinangle = sin angle in
        [| dir.(0) *. ilength *. sinangle;
           dir.(1) *. ilength *. sinangle;
           dir.(2) *. ilength *. sinangle;
           cos angle |]
    else [| 0.0; 0.0; 0.0; 1.0 |];;
    
let quaternionCopy q =
    [| q.(0); q.(1); q.(2); q.(3) |];;

let quaternionMultiply q1 q2 =
	let out = Array.make 16 0.0 in
    out.(0) <- q1.(3) *. q2.(0) +. q1.(0) *. q2.(0) +. q1.(1) *. q2.(2) -. q1.(2) *. q2.(1);
    out.(1) <- q1.(3) *. q2.(1) +. q1.(1) *. q2.(3) +. q1.(2) *. q2.(0) -. q1.(0) *. q2.(2);
    out.(2) <- q1.(3) *. q2.(2) +. q1.(2) *. q2.(3) +. q1.(0) *. q2.(1) -. q1.(1) *. q2.(0);
    out.(3) <- q1.(3) *. q2.(3) -. q1.(0) *. q2.(0) -. q1.(1) *. q2.(1) -. q1.(2) *. q2.(2);
	out;;

let quaternionToMatrix q =
    let x2 = q.(0) +. q.(0)
    and y2 = q.(1) +. q.(1)
    and z2 = q.(2) +. q.(2) in
    let xx = q.(0) *. x2
    and yy = q.(1) *. y2
    and zz = q.(2) *. z2
    and xy = q.(0) *. y2
    and yz = q.(1) *. z2
    and xz = q.(2) *. x2
    and wx = q.(3) *. x2
    and wy = q.(3) *. y2
    and wz = q.(3) *. z2 in
    let out = Array.make 16 0.0 in
    out.(0) <- 1.0 -. (yy +. zz);
    out.(1) <- xy +. wz;
    out.(2) <- xz -. wy;
    out.(3) <- 0.0;
    out.(4) <- xy -. wz;
    out.(5) <- 1.0 -. (xx +. zz);
    out.(6) <- yz +. wx;
    out.(7) <- 0.0;
    out.(8) <- xz +. wy;
    out.(9) <- yz -. wx;
    out.(10) <- 1.0 -. (xx +. yy);
    out.(11) <- 0.0;
    out.(12) <- 0.0;
    out.(13) <- 0.0;
    out.(14) <- 0.0;
    out.(15) <- 1.0;
	out;;
	